# -*- coding: utf-8 -*-

import json
import os
import re
import shutil
from copy import deepcopy
from importlib import import_module
from pathlib import Path
from typing import Optional, Union

import numpy as np
from monty.io import zopen
from monty.json import MontyEncoder
from monty.serialization import loadfn
from pymatgen.core.structure import Structure
from pymatgen.io.vasp.sets import VaspInputSet, Poscar, Potcar, Kpoints
from pymatgen.io.vasp.sets import (
    get_vasprun_outcar, get_structure_from_prev_run)
from vise.analyzer.band_gap import band_gap_properties
from vise.config import (
    DOS_STEP_SIZE, KPT_DENSITY, ENCUT_FACTOR_STR_OPT, ANGLE_TOL,
    SYMMETRY_TOLERANCE, BAND_REF_DIST, DEFAULT_NUM_NODES)
from vise.input_set.incar import ViseIncar
from vise.input_set.settings_incar import (
    TaskIncarSettings, XcIncarSettings, XcTaskIncarSettings,
    CommonIncarSettings)
from vise.input_set.settings_potcar import XcTaskPotcar
from vise.input_set.settings_structure_kpoints import TaskStructureKpoints
from vise.input_set.task import Task
from vise.input_set.xc import Xc
from vise.util.logger import get_logger
from vise.util.structure_handler import get_symbol_list

logger = get_logger(__name__)

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"


class ViseInputSet(VaspInputSet):
    """Vise version of VaspInputSet

    Use the kwargs for the options that control vasp input set as the dist
    can be easily updated. Here, a list of options are listed as below.

    Options:
        xc (Xc):
            Exchange-correlation (xc) defined in Xc.
        task (Task):
            Task defined in Task.
        sort_structure (bool):
            Whether to sort the elements using get_sorted_structure method
            of Structure class in pymatgen.
        standardize_structure (bool):
            Whether to convert the structure to a standardized primitive.
        kpt_mode (str):
            A string representing the k-point style that is used for the
            make_kpoints function. Now, "band", "primitive_uniform" and
            "manual_set" are supported. See make_kpoints docstring.
        kpt_density (float):
            K-point density in Angstrom along each direction used for
            determining k-point uniform mesh.
        kpt_shift (list):
            K-point shift in vasp definition.
        only_even (bool):
            Whether to ceil the numbers of kpoints to be even.
        band_ref_dist (float):
            A reference target distance between neighboring k-points in the
            path, in units of 1/Angstrom. The actual value will be as close
            as possible to this value, to have an integer number of points
            in each path. Note that seekpath default is 0.025.
        factor (int):
            Factor to be multiplied to the k-points along all the
            directions. This parameter should be distinguished from the
            kpt_density as it sets NKRED = factor for hybrid calculations:w
        symprec (float):
            Distance precision in Angstrom used for symmetry analysis.
        angle_tolerance (float):
            Angle precision in degrees used for symmetry analysis.
        charge (float):
            Charge state used for e.g., defect calculations.
        default_potcar (str / None):
            Default potcar name defined in potcar_set.yaml.
            By default, use "normal", but one needs to use GW potcars for GW
            calculation, which will be supported in the future.
        override_potcar_set (dict /None):
            Dict representation for overriding the potcar set, e.g.,
            {"Mg": "Mg_pv", "O": "O_h"}
        vbm_cbm (list):
            The absolute valence band maximum (vbm) and conduction band minimum
            (cbm) with [vbm, cbm] format in eV, which are used for determining
            ISMEAR and the spectra region of the dos and absorption spectra.
        aexx (float):
            Exchange mixing parameter, namely AEXX in INCAR for hybrid
            functional calculations.
        hubbard_u (bool):
            Whether to add Hubbard U parameter.
            By default (None), it is switched on for LDA and GGA calculations,
            while switched off for hybrid functional and GW approximations.
            Default U-parameters are determined in u_parameter_set.yaml.
        ldauu (dict):
            Set when users want to modify U parameters from default. E.g.,
            {"Ti": 4, "V": 3, "O": 5}
        ldaul (dict):
            Set when users want to modify orbitals from default.
            If ldauu add elements that are not set in u_parameter_set.yaml,
            ldaul also needs to be supplied for these elements. E.g.,
            {"Ti": 2, "V": 2, "O": 1}
        ldaul_set_name:
            LDAUU and LDAUL default set name defined in u_parameter_set.yaml.
        npar_kpar (bool):
            Whether to automatically set the NPAR and KPAR tags.
        encut (float):
            Cutoff energy in eV.
            This argument is useful when performing a set of calculations
            for models with different elements yet their energies to be
            compared, e.g., defect calculations with impurity.
        is_magnetization (bool):
            Whether to have magnetization. If exists, set
            ISPIN = 2 and time_reversal = False, the latter of which is
            necessary for band structure calculations.
        num_nodes (int):
            Numbers of nodes used to determine KPAR and NPAR INCAR setting.
        structure_opt_encut_factor (float):
            ENCUT multiplied factor for structure optimization, where encut
            needs to be increased to reduce the Pulay Stress errors.
    """
    GENERAL_OPTIONS = {"sort_structure": True,
                       "standardize_structure": True}

    TASK_OPTIONS = {"kpt_mode": "primitive_uniform",
                    "kpt_density": KPT_DENSITY,
                    "kpt_shift": None,
                    "only_even": False,
                    "band_ref_dist": BAND_REF_DIST,
                    "factor": None,
                    "symprec": SYMMETRY_TOLERANCE,
                    "angle_tolerance": ANGLE_TOL,
                    "dos_step_size": DOS_STEP_SIZE,
                    "charge": 0}

    XC_OPTIONS = {"potcar_set_name": "normal",
                  "override_potcar_set": None,
                  "vbm_cbm": None,
                  "aexx": 0.25,
                  "hubbard_u": None,
                  "ldauu": None,
                  "ldaul": None,
                  "ldaul_set_name": "default"}

    XC_TASK_OPTIONS = {"npar_kpar": True,
                       "encut": None}

    COMMON_OPTIONS = {"is_magnetization": False,
                      "num_nodes": DEFAULT_NUM_NODES,
                      "structure_opt_encut_factor": ENCUT_FACTOR_STR_OPT}

    ALL_OPTIONS = {**GENERAL_OPTIONS, **TASK_OPTIONS, **XC_OPTIONS,
                   **XC_TASK_OPTIONS, **COMMON_OPTIONS}

    def __init__(self,
                 structure: Structure,
                 xc: Xc,
                 task: Task,
                 kpoints: Kpoints,
                 potcar: Potcar,
                 incar_settings: dict,
                 files_to_transfer: dict,
                 **kwargs):
        """
        Args:
            structure (Structure): The Structure to create inputs for.
            xc (Xc): Exchange-correlation (xc) defined in Xc.
            task (Task): Task defined in Task.
            kpoints (Kpoints): Kpoints class object
            potcar (Potcar): Potcar class object
            incar_settings (dict): INCAR settings generated by some options
            files_to_transfer (dict):
            kwargs (dict): OPTION arguments.
        """

        self.structure = structure
        self.xc = xc
        self.task = task
        self._kpoints = kpoints
        self._potcar = potcar
        self.incar_settings = incar_settings
        self.files_to_transfer = files_to_transfer
        self.kwargs = kwargs

    @classmethod
    def make_input(cls,
                   structure: Structure,
                   task: Union[str, Task] = "structure_opt",
                   xc: Union[str, Xc] = "pbe",
                   prev_set: "ViseInputSet" = None,
                   abs_files_to_transfer: Optional[dict] = None,
                   user_incar_settings: Optional[dict] = None,
                   **kwargs) -> "ViseInputSet":
        """Construct ViseInputSet from some options including xc and task.

        To make a simple but practical rule for inheriting the previous
        calculation condition and effectively use its results, we adopt
        fail-safe rule.

        When, the prev_set is set, we compare task and xc between current
        input and prev_set, and inherit some options depending on whether
        task and/or xc are common. For instance, when we calculate the band
        structure with the HSE06 hybrid functional, it would be better to
        generate the WAVECAR file with e.g., PBE functional to reduce the
        computational cost. Then, the initial task was done with task=Task.band
        and xc=Xc.pbe with user_incar_setting={"LWAVE": True}, and next task is
        performed with task=Task.band, xc=Xc.hs. Then, task is common, so
        input set related to TaskStructureKpoints and TaskIncarSettings are
        inherited. Note that, CommonIncarSettings options are always inherited.

        Other notes are as follows.

        Note1: Charge set to structure is ignored when determining NELECT. For
               this purpose, charge option needs to be used.
        Note2: When the structure is changed via find_spglib_standard_primitive,
               all the site properties are removed.
        Note3: When different version of ViseInputSet with different defaults,
               the consistency of input set is destroyed, so the same version
               must be used when options are inherited.
        Note4: Other INCAR flags than those defined by TaskIncarSettings,
               XcIncarSettings, XcTaskIncarSettings, and CommonIncarSettings are
               not inherited. When some of them need to be inherited, they
               should be added to COMMON_OPTIONAL_FLAGS.
        Note5: user_incar_settings is not inherited from prev_set. One needs
               to explicitly specify it, again, if needed.
        Note6: Since default args are used in main function, set
               "structure_opt" and "pbe" strings as defaults.

        Args:
            structure (Structure):
                The Structure to create inputs for.
            task (str/Task):
                Task defined in Task.
            xc (str/Xc):
                Exchange-correlation (xc) defined in Xc.
            prev_set (ViseInputSet):
                Previous ViseInputSet.
            abs_files_to_transfer (dict):
                Keys are file names with absolute paths to be transferred.
                Values mean the transfer modes, where there are three modes,
                "c": copy file, "m": move file, "l": make symbolic link
                e.g., {"/..../CHGCAR": "c", "/..../WAVECAR": "l", ..}
            user_incar_settings (dict):
                User INCAR settings.
                e.g., {"NSW": 100, "LWAVE": True, "LASPH": "A", ..}
            kwargs (dict):
                OPTION arguments.

        Return:
            ViseInputSet class object.
        """
        abs_files_to_transfer = abs_files_to_transfer or {}
        user_incar_settings = user_incar_settings or {}

        if isinstance(task, str):
            task = Task.from_string(task)
        if isinstance(xc, str):
            xc = Xc.from_string(xc)

        # First, set default.
        opts = deepcopy(cls.ALL_OPTIONS)

        # Second, override with previous condition
        if prev_set:
            key_set = set(cls.COMMON_OPTIONS.keys())
            if prev_set.task == task:
                key_set.update(cls.TASK_OPTIONS.keys())
            if prev_set.xc == xc:
                key_set.update(cls.XC_OPTIONS.keys())
            if prev_set.task == task and prev_set.xc == xc:
                key_set.update(cls.XC_TASK_OPTIONS)

            for k in key_set:
                opts[k] = prev_set.kwargs[k]

        # Third, override with keyword arguments
        for k in kwargs:
            if k not in opts.keys():
                logger.warning(f"Keyword {k} is not adequate for ViseInputSet.")
        opts.update(kwargs)

        task_str_kpt = TaskStructureKpoints.from_options(
            task=task,
            original_structure=structure,
            standardize_structure=opts["standardize_structure"],
            sort_structure=opts["sort_structure"],
            is_magnetization=opts["is_magnetization"],
            kpt_mode=opts["kpt_mode"],
            kpt_density=opts["kpt_density"],
            kpt_shift=opts["kpt_shift"],
            only_even=opts["only_even"],
            band_ref_dist=opts["band_ref_dist"],
            factor=opts["factor"],
            symprec=opts["symprec"],
            angle_tolerance=opts["angle_tolerance"])

        orig_matrix = structure.lattice.matrix
        matrix = task_str_kpt.structure.lattice.matrix
        structure_changed = \
            not np.allclose(orig_matrix, matrix, atol=opts["symprec"])

        if structure_changed and opts["charge"] != 0:
            raise ValueError("Structure is changed but charge is set.")

        if structure_changed:
            # The following files are useless when lattice is changed.
            pattern = \
                re.compile("|".join([r"CHGCAR$", r"WAVECAR$", r"WAVEDER"]))
            for f in abs_files_to_transfer:
                if pattern.match(f):
                    abs_files_to_transfer.pop(f, None)

        # unique_justseen https://docs.python.org/ja/3/library/itertools.html
        # ["H", "H", "O", "O", "H"] -> ['H', 'O', 'H']
        symbol_list = get_symbol_list(task_str_kpt.structure)

        xc_task_potcar = XcTaskPotcar.from_options(
            xc=xc,
            symbol_list=symbol_list,
            potcar_set_name=opts["potcar_set_name"],
            override_potcar_set=opts["override_potcar_set"])

        task_settings = TaskIncarSettings.from_options(
            task=task,
            structure=task_str_kpt.structure,
            potcar=xc_task_potcar.potcar,
            num_kpoints=task_str_kpt.num_kpts,
            max_enmax=xc_task_potcar.max_enmax,
            is_magnetization=opts["is_magnetization"],
            vbm_cbm=opts["vbm_cbm"],
            npar_kpar=opts["npar_kpar"],
            num_nodes=opts["num_nodes"],
            encut=opts["encut"],
            structure_opt_encut_factor=opts["structure_opt_encut_factor"],
            dos_step_size=opts["dos_step_size"])

        xc_settings = XcIncarSettings.from_options(
            xc=xc,
            symbol_list=symbol_list,
            factor=task_str_kpt.factor,
            aexx=opts["aexx"],
            hubbard_u=opts["hubbard_u"],
            ldauu=opts["ldauu"],
            ldaul=opts["ldaul"],
            ldaul_set_name=opts["ldaul_set_name"])

        xc_task_settings = XcTaskIncarSettings.from_options()

        common_settings = CommonIncarSettings.from_options(
            potcar=xc_task_potcar.potcar,
            composition=task_str_kpt.structure.composition,
            charge=opts["charge"])

        incar_settings = \
            {**task_settings.settings, **xc_settings.settings,
             **xc_task_settings.settings, **common_settings.settings}

        # user_incar_settings is the top prioritized.
        incar_settings.update(user_incar_settings)

        # TODO: tweak the unfavorable combination of the input set.
        # e.g., Avoiding ICHARG = 11 is a must for hybrid functional.

        return cls(structure=task_str_kpt.structure,
                   task=task,
                   xc=xc,
                   kpoints=task_str_kpt.kpoints,
                   potcar=xc_task_potcar.potcar,
                   incar_settings=incar_settings,
                   files_to_transfer=abs_files_to_transfer, **opts)

    @property
    def potcar(self):
        return self._potcar

    @property
    def kpoints(self):
        return self._kpoints

    @property
    def incar(self):
        incar = ViseIncar.from_dict(self.incar_settings)
        return incar

    @property
    def poscar(self):
        return Poscar(self.structure)

    def write_input(self,
                    output_dir: str,
                    make_dir_if_not_present: bool = True,
                    include_cif: bool = False,
                    to_json_file: bool = True,
                    json_filename: str = "vise.json") -> None:
        """Write vasp input files and handle transferred files.

        Args:
            See docstring of write_input of VaspInputSet

        Return:
            None
        """
        super().write_input(output_dir=output_dir,
                            make_dir_if_not_present=make_dir_if_not_present,
                            include_cif=include_cif)

        # Overwrite only POSCAR to increase the digit for frac coords.
        poscar_file = Path(output_dir) / "POSCAR"
        self.poscar.write_file(str(poscar_file), significant_figures=10)

        out_dir = Path(output_dir).absolute()
        for key, value in self.files_to_transfer.items():
            try:
                # full path is usually safer for symbolic link.
                filepath = Path(key).absolute()
                name = filepath.name
                if value == "c" or value == "m":
                    with zopen(filepath, "rb") as fin, \
                            zopen((out_dir / name), "wb") as fout:
                        if value == "c":
                            shutil.copyfileobj(fin, fout)
                        elif value == "m":
                            shutil.move(fin, fout)
                elif value == "l":
                    os.symlink(filepath, out_dir / name)

            except FileNotFoundError:
                logger.warning(f"{key} does not exist.")

        if to_json_file:
            self.to_json_file(json_filename)

    def as_dict(self, **kwargs):

        parent_module = self.__class__.__module__.split('.')[0]
        module_version = import_module(parent_module).__version__
        # Xc and Task objects must be converted to string for to_json_file as
        # Enum is not compatible with MSONable.
        d = {"@module":             self.__class__.__module__,
             "@class":              self.__class__.__name__,
             "@version":            module_version,
             "structure":           self.structure,
             "xc":                  str(self.xc),
             "task":                str(self.task),
             "kpoints":             self.kpoints,
             "potcar":              self.potcar.as_dict(),
             "incar_settings":      self.incar_settings,
             "files_to_transfer":   self.files_to_transfer,
             "kwargs":              self.kwargs}

        return d

    def __eq__(self, other):
        try:
            return self.as_dict() == other.as_dict()
        except AttributeError:
            return False

    def __repr__(self):
        out = [f"task: {self.task}",
               f"xc: {self.xc}"]
        return "\n".join(out)

    def __str__(self):
        return self.__repr__()

    @classmethod
    def from_dict(cls, d):
        # Programmatic access to enumeration members in Enum class.
        structure = d["structure"]
        if isinstance(structure, dict):
            structure = Structure.from_dict(structure)

        return cls(structure=structure,
                   xc=Xc.from_string(d["xc"]),
                   task=Task.from_string(d["task"]),
                   kpoints=d["kpoints"],
                   potcar=Potcar.from_dict(d["potcar"]),
                   incar_settings=d["incar_settings"],
                   files_to_transfer=d["files_to_transfer"],
                   **d["kwargs"])

    @classmethod
    def load_json(cls, filename: str = "vise.json"):
        return loadfn(filename)

    def to_json_file(self, filename: str = "vise.json"):
        with open(filename, 'w') as fw:
            json.dump(self.as_dict(), fw, indent=2, cls=MontyEncoder)

    @classmethod
    def from_prev_calc(cls,
                       dirname: str,
                       task: Optional[Task] = None,
                       xc: Optional[Xc] = None,
                       json_filename: str = "vise.json",
                       parse_calc_results: bool = True,
                       parse_incar: bool = True,
                       sort_structure: bool = True,
                       standardize_structure: bool = False,
                       files_to_transfer: Optional[dict] = None,
                       user_incar_settings: Optional[dict] = None,
                       contcar_filename: str = "CONTCAR",
                       **kwargs) -> "ViseInputSet":
        """Constructor based on the previous calculations.

        If the task and/or xc are not set, the previous ones are assumed.

        Args:
            dirname (str):
                Directory name to be parsed.
            xc (Xc):
                Exchange-correlation (xc) defined in Xc.
            task (Task):
                Task defined in Task.
            json_filename (str):
                Json filename including ViseInputSet data.
            parse_calc_results (bool):
                Whether to parse the previous calculation results for band
                edges, magnetization.
            parse_incar (bool):
                Whether to parse the previous INCAR files, which could include
                other options that are not determined in vise.
            sort_structure (bool):
                Whether to sort the elements using get_sorted_structure method
                of Structure class in pymatgen.
            standardize_structure (bool):
                Whether to convert the structure to a standardized primitive.
            files_to_transfer (dict):
                Keys are file names to be transferred in the directory.
                Values mean the transfer modes, where there are three modes,
                "c": copy file, "m": move file, "l": make symbolic link
                e.g., {"CHGCAR": "c", "WAVECAR": "l", ..}
            user_incar_settings (dict):
                User INCAR settings.
                e.g., {"NSW": 100, "LWAVE": True, "LASPH": "A", ..}
            contcar_filename (str):
                CONTCAR type filename. Used if the user change the name.
            kwargs:
                Other OPTION arguments.

        Return:
            ViseInputSet class object.
        """
        kwargs = kwargs or {}
        kwargs["sort_structure"] = sort_structure
        kwargs["standardize_structure"] = standardize_structure

        path = Path(dirname)

        input_set = ViseInputSet.load_json(path / json_filename)

        if parse_calc_results:
            vasprun, outcar = get_vasprun_outcar(dirname)
            structure = get_structure_from_prev_run(vasprun, outcar)

            gap_properties = band_gap_properties(vasprun, outcar)
            if gap_properties:
                _, vbm, cbm = gap_properties
                kwargs["vbm_cbm"] = [vbm["energy"], cbm["energy"]]

            if "magmom" in structure.site_properties.keys():
                abs_max_mag = abs(max(structure.site_properties["magmom"],
                                      key=abs))
            else:
                abs_max_mag = 0
            if vasprun.is_spin and abs_max_mag > 0.05:
                kwargs["is_magnetization"] = True

        else:
            contcar = path / contcar_filename
            poscar = path / "POSCAR"
            if contcar.is_file and os.stat(contcar).st_size != 0:
                logger.info(f"{contcar} is parsed for structure.")
                structure = Structure.from_file(contcar)
            elif poscar.is_file and os.stat(poscar).st_size != 0:
                logger.warning(f"{poscar} is parsed for structure.")
                structure = Structure.from_file(poscar)
            else:
                raise FileNotFoundError("CONTCAR or POSCAR does not exist.")

        if parse_incar:
            input_set.incar_settings = ViseIncar.from_file(path / "INCAR")

        if files_to_transfer:
            abs_files_to_transfer = {}
            for filename, value in files_to_transfer.items():
                f = path / filename
                if not f.is_file():
                    logger.warning(f"{f} does not exist.")
                elif os.stat(f).st_size == 0:
                    logger.warning(f"{f} is empty.")
                elif value[0].lower() not in ["l", "c", "m"]:
                    logger.warning(f"{value} option for {filename} is invalid.")
                else:
                    abs_files_to_transfer[str(f)] = value[0].lower()
            files_to_transfer = abs_files_to_transfer

        return cls.make_input(structure=structure,
                              task=task or input_set.task,
                              xc=xc or input_set.xc,
                              prev_set=input_set,
                              abs_files_to_transfer=files_to_transfer,
                              user_incar_settings=user_incar_settings,
                              **kwargs)





# !/usr/bin/env python
# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from vise.cli.tests.mymodule import RemovalService, UploadService

import unittest
from unittest import mock


class RemovalServiceTestCase(unittest.TestCase):

    @mock.patch('vise.cli.tests.mymodule.os.path')
    @mock.patch('vise.cli.tests.mymodule.os')
    def test_rm(self, mock_os, mock_path):
        # instantiate our service
        reference = RemovalService()

        # set up the mock
        mock_path.isfile.return_value = False

        reference.rm("any path")

        # test that the remove call was NOT called.
        self.assertFalse(mock_os.remove.called,
                         "Failed to not remove the file if not present.")

        # make the file 'exist'
        mock_path.isfile.return_value = True

        reference.rm("any path")

        mock_os.remove.assert_called_with("any path")


class UploadServiceTestCase(unittest.TestCase):

    @mock.patch.object(RemovalService, 'rm')
    def test_upload_complete(self, mock_rm):
        # build our dependencies
        removal_service = RemovalService()
        reference = UploadService(removal_service)

        # call upload_complete, which should, in turn, call `rm`:
        reference.upload_complete("my uploaded file")

        # check that it called the rm method of any RemovalService
        mock_rm.assert_called_with("my uploaded file")

        # check that it called the rm method of _our_ removal_service
        removal_service.rm.assert_called_with("my uploaded file")


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np
from pathlib import Path
from pymatgen.io.vasp.outputs import Vasprun, Outcar
from pymatgen.electronic_structure.core import Spin

from vise.analyzer.band_gap import band_gap_properties, edge_info
from vise.util.testing import ViseTest


parent_dir = Path(__file__).parent


class BandGapPropertiesTest(ViseTest):
    def test_mgo(self):
        vasprun = Vasprun(parent_dir / "MgO_band_vasprun.xml")
        outcar = Outcar(parent_dir / "MgO_band_OUTCAR")
        expected = ({'energy': 4.6597,
                     'direct': True},
                    {'energy': 3.0663,
                     'spin': None,
                     'band_index': 3,
                     'kpoints': [0.0, 0.0, 0.0]},
                    {'energy': 7.726,
                     'spin': None,
                     'band_index': 4,
                     'kpoints': [0.0, 0.0, 0.0]})
        self.assertEqual(expected, band_gap_properties(vasprun, outcar))

    def test_bata6o16(self):
        vasprun = Vasprun(parent_dir / "BaTa6O16_band_vasprun.xml")
        outcar = Outcar(parent_dir / "BaTa6O16_band_OUTCAR")
        expected = ({'energy': 2.6454000000000004,
                     'direct': False},
                    {'energy': 1.8945000000000001,
                     'spin': None,
                     'band_index': 67,
                     'kpoints': [-0.5, 0.5, 0.46774194]},
                    {'energy': 4.5399000000000003,
                     'spin': None,
                     'band_index': 68,
                     'kpoints': [0.0, 0.0, 0.5]})
        self.assertEqual(expected, band_gap_properties(vasprun, outcar))

    def test_mno(self):
        vasprun = Vasprun(parent_dir / "MnO_uniform_vasprun.xml")
        outcar = Outcar(parent_dir / "MnO_uniform_OUTCAR")
        expected = ({'energy': 0.47020000000000017,
                     'direct': False},
                    {'energy': 4.6665999999999999,
                     'spin': 1,
                     'band_index': 8,
                     'kpoints': [0.42857143, 0.0, 0.0]},
                    {'energy': 5.1368,
                     'spin': 1,
                     'band_index': 9,
                     'kpoints': [0.0, 0.0, 0.0]})
        self.assertEqual(expected, band_gap_properties(vasprun, outcar))


class EdgeInfoTest(ViseTest):
    def setUp(self) -> None:
        values = np.array([[[1.0, 2.0], [3.0, 4.0]], [[5.0, 6.0], [7.0, 8.0]]])
        # [[[ 1.  2.]   1st k: 1st band (hob)
        #   [ 3.  4.]]  1st k: 2nd band (lub)
        #
        #  [[ 5.  6.]   2nd k: 1st band (hob)
        #   [ 7.  8.]]] 2nd k: 2st band (lub)
        self.eigenvalues = {Spin.up: values}

    def test(self):
        actual = edge_info(self.eigenvalues,
                           hob_index=0,
                           spin=Spin.up)
        expected = 5.0, 1, 3.0, 0
        self.assertEqual(expected, actual)






# -*- coding: utf-8 -*-
from math import ceil
from pathlib import Path
from typing import Optional
from typing import Tuple

from monty.serialization import loadfn
from pymatgen import Composition
from pymatgen.io.vasp import Potcar
from vise.input_set.datasets.element_parameters import unoccupied_bands
from vise.util.logger import get_logger

logger = get_logger(__name__)
SET_DIR = Path(__file__).parent / "datasets"


def load_potcar_yaml(set_name: Optional[str] = "normal",
                     override_potcar_set: Optional[dict] = None) -> dict:
    """Load the yaml setting files for config and POTCAR list.

    Args:
        set_name (str):
            Potcar set name.
        override_potcar_set (dict):
            User specifying POTCAR set

    Returns:
          Dictionary of potcar_set, like {"Zr": "Zr_pv", ...}
    """
    potcar_set = loadfn(SET_DIR / "potcar_set.yaml")
    try:
        potcar = potcar_set[set_name]
    except KeyError:
        logger.critical(f"The accepted potcar set name is {potcar_set.keys()}")
        raise

    if override_potcar_set:
        potcar.update(override_potcar_set)

    return potcar


def load_default_incar_settings(yaml_filename: str,
                                required_flags: set,
                                optional_flags: set,
                                key_name: str) -> dict:
    """Load the yaml setting files for config and POTCAR list.

    Args:
        yaml_filename (str):
            Yaml filename with xxx_incar_set.yaml.
        required_flags (set):
            INCAR flags that must exist in the dict as a value of key_name key
            in yaml file.
        optional_flags (set):
            INCAR flags that that could exist in the dict as a value of key_name
            key in yaml file.
        key_name (str):
            Key name such as "structure_opt" or "hse".

    Returns:
          settings (dict):
    """
    incar_set = loadfn(SET_DIR / yaml_filename)
    settings = {}
    for f in required_flags:
        settings[f] = incar_set[f][key_name]

    for f in optional_flags:
        try:
            settings[f] = incar_set[f][key_name]
        except KeyError:
            continue

    return settings


def check_keys(d: dict, required: set, optional: set) -> bool:
    """Check if the required keys exist and other keys are included in optional

    Args:
        d (dict):
            Dictionary to be checked.
        required (set):
            Required key set.
        optional (set):
            Optional key set.

    Returns:
        Simply raise KeyError if the condition is not satisfied.
        If succeed, True is returned.
    """
    check_required = set(required)
    for key in d:
        check_required.discard(key)
        if key not in required | optional:
            raise KeyError(f"{key} does not belong.")

    if check_required:
        raise KeyError(f"{check_required} must exist in {d}.")

    return True


def nelect(composition: Composition, potcar: Potcar, charge: int = 0) -> int:
    """Gets the default number of electrons for a given composition. """
    num_nelect = 0
    for pt in potcar:
        num_nelect += composition.element_composition[pt.element] * pt.ZVAL

    return int(num_nelect) - charge


def nbands(composition: Composition, potcar: Potcar) -> int:
    """Calculate total number of bands for unoccupied related properties

    Useful for optical absorption, band structure, and DOS.

    Args:
        composition (Composition): Input composition
        potcar (Potcar): Input Potcar object

    Returns:
        Int of number of bands
    """
    num_bands = \
        sum([composition[c] * (p.nelectrons / 2 + unoccupied_bands[str(c)])
             for c, p in zip(composition, potcar)])

    return ceil(num_bands)


def calc_npar_kpar(num_kpoints: int, num_nodes: int) -> Tuple[int, int]:
    """Recommend suitable NPAR and KPAR params in INCAR from num of kpts.

    The test is performed using 36 core node.

    Args:
        num_kpoints (int): Number of irreducible k-points.
        num_nodes (int): Number of nodes.

    Returns:
        Tuple of recommended NPAR and KPAR.
    """

    kpar_set = {
        1:    [1, 1, 1],
        2:    [2, 2, 2],
        3:    [3, 3, 3],
        4:    [2, 4, 4],
        5:    [1, 2, 2],
        6:    [3, 6, 6],
        7:    [2, 2, 4],
        8:    [4, 8, 8],
        9:    [3, 3, 3],
        10:   [2, 2, 2],
        11:   [2, 2, 2],
        12:   [3, 4, 4],
        13:   [2, 2, 2],
        14:   [2, 2, 2],
        15:   [3, 2, 2],
        16:   [4, 4, 8],
        17:   [2, 4, 4],
        18:   [3, 6, 4],
        19:   [2, 4, 4],
        20:   [2, 4, 4],
        21:   [3, 3, 3],
        22:   [2, 4, 4],
        23:   [2, 4, 4],
        24:   [3, 6, 8],
        25:   [2, 4, 4],
        26:   [2, 4, 4],
        27:   [3, 4, 4],
        28:   [2, 4, 4],
        29:   [2, 4, 4],
        30:   [3, 6, 4],
        31:   [3, 4, 4],
        32:   [3, 4, 8],
        33:   [3, 4, 4],
        34:   [3, 4, 4],
        35:   [3, 4, 4],
        36:   [4, 8, 16],
        48:   [4, 8, 16],
        60:   [4, 8, 16],
        72:   [4, 8, 16],
        None: [4, 8, 16]}

    num_kpt_key = num_kpoints if num_kpoints in kpar_set else None

    if num_nodes == 2:
        kpar = kpar_set[num_kpt_key][1]
    elif num_nodes == 4:
        kpar = kpar_set[num_kpt_key][2]
    else:
        kpar = kpar_set[num_kpt_key][0]

    if kpar == 1:
        npar = 2
    else:
        npar = 1

    return kpar, npar



# -*- coding: utf-8 -*-

from itertools import groupby
import operator
from typing import Tuple

import seekpath
import spglib

from pymatgen import Structure
from pymatgen.core.periodic_table import Element

from vise.config import ANGLE_TOL, SYMMETRY_TOLERANCE, BAND_REF_DIST
from vise.util.logger import get_logger

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"

logger = get_logger(__name__)


def get_symmetry_dataset(structure: Structure,
                         symprec: float = SYMMETRY_TOLERANCE,
                         angle_tolerance: float = ANGLE_TOL) -> dict:
    """ Get spglib symmetry dataset from a Structure.

    Args:
        structure (Structure):
            Pymatgen Structure class object
        symprec (float):
            Distance tolerance in cartesian coordinates Unit is compatible with
            the cell.
        angle_tolerance (float):
            Angle tolerance used for symmetry analyzer.

    Return:
        spglib symmetry dataset. See docstrings of spglib for details.
    """
    cell = structure_to_spglib_cell(structure)
    return spglib.get_symmetry_dataset(cell=cell,
                                       symprec=symprec,
                                       angle_tolerance=angle_tolerance)


def structure_to_spglib_cell(structure: Structure) -> Tuple[list, list, list]:
    """Convert structure to spglib cell tuple.

    Return a cell tuple used by spglib that is composed of lattice parameters,
    atomic positions in fractional coordinates, and atomic numbers.

    Args:
        structure (Structure):
            Pymatgen Structure class object

    Returns:
        Tuple of (lattice matrix, fractional coordinates, atomic numbers
    """
    lattice = list(structure.lattice.matrix)
    positions = structure.frac_coords.tolist()
    atomic_numbers = [i.specie.number for i in structure.sites]

    return lattice, positions, atomic_numbers


def spglib_cell_to_structure(cell: tuple) -> Structure:
    """
    Args:
        cell (3 tuple):
            Lattice parameters, atomic positions in fractional coordinates,
            and corresponding atom numbers
    Returns:
         pymatgen Structure class object from spglib cell tuple.
    """
    species = [Element.from_Z(i) for i in cell[2]]
    return Structure(cell[0], species, cell[1])


def find_spglib_standard_conventional(structure: Structure,
                                      symprec: float = SYMMETRY_TOLERANCE,
                                      angle_tolerance: float = ANGLE_TOL
                                      ) -> Structure:
    """

    Args:
        structure (Structure):
            Pymatgen Structure class object
        symprec (float):
            Distance tolerance in cartesian coordinates Unit is compatible with
            the cell.
        angle_tolerance (float):
            Angle tolerance used for symmetry analyzer.
    Returns:
        Structure of a standard conventional unit cell.
    """
    cell = structure_to_spglib_cell(structure)
    return spglib_cell_to_structure(
        spglib.standardize_cell(cell=cell,
                                to_primitive=False,
                                no_idealize=False,
                                symprec=symprec,
                                angle_tolerance=angle_tolerance))


def find_spglib_primitive(structure: Structure,
                          symprec: float = SYMMETRY_TOLERANCE,
                          angle_tolerance: float = ANGLE_TOL
                          ) -> Tuple[Structure, bool]:
    """
    Args:
        structure (Structure):
            Pymatgen Structure class object
        symprec (float):
            Distance tolerance in cartesian coordinates Unit is compatible with
            the cell.
        angle_tolerance (float):
            Angle tolerance used for symmetry analyzer.

    Returns:
        Structure of a primitive unit cell.
    """
    cell = structure_to_spglib_cell(structure)
    primitive_cell = spglib.find_primitive(cell=cell,
                                           symprec=symprec,
                                           angle_tolerance=angle_tolerance)
    primitive_structure = spglib_cell_to_structure(primitive_cell)
    is_structure_changed = structure.lattice != primitive_structure.lattice

    return primitive_structure, is_structure_changed


def find_hpkot_primitive(structure: Structure,
                         symprec: float = SYMMETRY_TOLERANCE,
                         angle_tolerance: float = ANGLE_TOL
                         ) -> Structure:
    """
    Args:
        structure (Structure):
            Pymatgen Structure class object
        symprec (float):
            Distance tolerance in cartesian coordinates Unit is compatible with
            the cell.
        angle_tolerance (float):
            Angle tolerance used for symmetry analyzer.

    Returns:
         Structure of a hpkot primitive unit cell.
    """
    cell = structure_to_spglib_cell(structure)
    res = seekpath.get_explicit_k_path(structure=cell,
                                       symprec=symprec,
                                       angle_tolerance=angle_tolerance)

    return seekpath_to_hpkot_structure(res)


def structure_to_seekpath(structure: Structure,
                          time_reversal: bool = True,
                          ref_distance: float = BAND_REF_DIST,
                          recipe: str = 'hpkot',
                          threshold: float = 1e-7,
                          symprec: float = SYMMETRY_TOLERANCE,
                          angle_tolerance: float = ANGLE_TOL) -> dict:
    """Get full information of seekpath band path.

    Args:
        structure (Structure):
            Pymatgen Structure class object
        time_reversal (bool):
            If the time reversal symmetry exists
        ref_distance (float):
            Distance for the k-point mesh.
        recipe (str):
            See docstrings of seekpath.
        threshold (float):
            To use to verify if we are in edge case (see seekpath).
        symprec (float):
            Distance tolerance in cartesian coordinates Unit is compatible with
            the cell.
        angle_tolerance (float):
            Angle tolerance used for symmetry analyzer.

    Return:
        Dict with some properties. See docstrings of seekpath.
    """
    cell = structure_to_spglib_cell(structure)
    res = seekpath.get_explicit_k_path(cell,
                                       with_time_reversal=time_reversal,
                                       reference_distance=ref_distance,
                                       recipe=recipe,
                                       threshold=threshold,
                                       symprec=symprec,
                                       angle_tolerance=angle_tolerance)

    # If numpy.allclose is too strict in pymatgen.core.lattice __eq__,
    # make almost_equal
    if structure.lattice != seekpath_to_hpkot_structure(res).lattice:
        logger.warning(
            "Given structure is modified to be compatible with HPKOT k-path.")

    return res


def seekpath_to_hpkot_structure(res: dict) -> Structure:
    """
    Args:
        res (dict):
            seekpath res dictionary.

    Returns:
         pymatgen Structure obtained from seekpath res dictionary.
    """
    lattice = res["primitive_lattice"]
    element_types = res["primitive_types"]
    species = [Element.from_Z(i) for i in element_types]
    positions = res["primitive_positions"]

    return Structure(lattice, species, positions)


def get_symbol_list(structure: Structure):
    species = [str(s) for s in structure.species]
    # unique_justseen https://docs.python.org/ja/3/library/itertools.html
    # ["H", "H", "O", "O", "H"] -> ['H', 'O', 'H']
    return list(map(next, map(operator.itemgetter(1), groupby(species, None))))



# -*- coding: utf-8 -*-
from pathlib import Path

from pymatgen.core.structure import Structure
from pymatgen.util.testing import PymatgenTest


class ViseTest(PymatgenTest):
    """Extends PymatgenTest with some path modification. """
    MODULE_DIR = Path(__file__).absolute().parent
    TEST_FILES_DIR = MODULE_DIR / ".." / "test_files"
    POSCARS_DIR = TEST_FILES_DIR / "poscars"

    @classmethod
    def get_structure_by_name(cls, name: str) -> Structure:
        filename = cls.POSCARS_DIR / ("POSCAR-" + name)
        return Structure.from_file(filename)

    @classmethod
    def get_structure_by_sg(cls, sg: int) -> Structure:
        filename = cls.POSCARS_DIR / "poscar_by_sg" / ("POSCAR_" + str(sg))
        return Structure.from_file(filename)

    @classmethod
    def get_filename(cls, name) -> Path:
        return cls.TEST_FILES_DIR / name
